import os
from pathlib import Path
from pprint import pprint
import yaml
import re
import time
from shutil import copy2

# loop through every file in the dir
root_path = '../../webwork-open-problem-library/Contrib/BrockPhysics/College_Physics_Urone/'
root_dest_folder = '../../instructor_physics_bank/source_ww/output_md/College_Physics_Urone/'

# variable declaration
counter = 0
source_files = []
src_dirs = []
title = topic = author = editor = date = source = template_version = problem_type = attribution = outcomes = difficulty = randomization = taxonomy = ""
tags = assets = altText = image_line = []
total_start_time = time.process_time()


# Variable declaration for Webwork keywords
metadata_end_src = "DOCUMENT();"
marcos_end_src = "TEXT(beginproblem());"
variables_start_src = "showHint"
problem_body_start_src = "BEGIN_TEXT"
problem_body_end_src = "END_TEXT"
hint_start_src = "BEGIN_HINT"
hint_end_src = "END_HINT"
ans_rule_src = "ans_rule"
ans_src = "ANS"
hint_src = "hint"
image_src = "image"
context_src = "Context"
partial_answer_src = "showPartialCorrectAnswers"

# extract file structure from source directory (handles ALL sub-directories)
# for loop runs based # of folders in src
for root, dirs, files in os.walk(root_path):
    for name in dirs:
        dest_folder = os.path.join(root, name).removeprefix(root_path)
        src_dirs.append(root_dest_folder + dest_folder)


def split_file(file_content):
    # TODO: once all functions are completed, convert global variables above into local variables
    # split the file into bite-size pieces to increase speed and reduce bugs
    metadata_content = file_content[:file_content.find(metadata_end_src)]
    macros = file_content[file_content.find(metadata_end_src):file_content.find(marcos_end_src)]
    question_variables = file_content[file_content.find(marcos_end_src):file_content.find(problem_body_start_src)]
    question_body = [file_content[file_content.find(problem_body_start_src):file_content.rfind(problem_body_end_src)]]
    question_ans = re.findall(r"ANS(\(.+?[?<!)]\));", file_content)
    question_hint = [file_content[file_content.find(hint_start_src):file_content.find(hint_end_src)]]
    question_solution = [file_content[file_content.find(marcos_end_src):file_content.find(problem_body_start_src)]]
    return {'metadata': metadata_content,
            'macros': macros,
            'question_variables': question_variables,
            'question_body': question_body,
            'question_ans': question_ans,
            'question_solution': question_solution,
            'question_hint': question_hint}


def metadata_extract(metadata_content):
    metadata = "## "
    chapter_src = "DBchapter"
    section_src = "DBsection"
    author_src = "AuthorText1"
    editor_src = "Editor"
    keywords_src = "KEYWORDS"
    date_src = "Date"

    title_ = topic_ = author_ = editor_ = tags_ = date_ = None

    # Look for keywords declared above and extract them from metadata content
    for item in metadata_content.split("\n"):
        if metadata + chapter_src in item:
            title_ = item[item.find("(") + 1:item.find(")")].replace("'", "")
        if metadata + section_src in item:
            topic_ = item[item.find("(") + 1:item.find(")")].replace("'", "")
        if metadata + author_src in item:
            author_ = item[item.find("(") + 1:item.find(")")].replace("'", "")
        if metadata + editor_src in item:
            editor_ = item[item.find("(") + 1:item.find(")")].replace("'", "")
        else:
            editor_ = 'N/A'
        if keywords_src in item:
            tags_ = item[12:-1].replace("'", "").replace('"', '').strip().split(',')
        if metadata + date_src in item:
            date_ = item[item.find("(") + 1:item.find(")")].replace("'", "")
    return {'title': title_,
            'topic': topic_,
            'author': author_,
            'editor': editor_,
            'tags': tags_,
            'date': date_}


def determine_problem_type(question_ans, filename):
    # determine what type of question is based on the ANS(type)
    numerical_type = "num_cmp"
    functional_type = "fun_cmp"
    checkbox_type = "checkbox_cmp"
    text_type = "str_cmp"
    question_type = []

    # extracts answer variable from ANS(num_cmp("variable"))
    answer_variable = re.findall(r'"\$([^"]*)"', str(question_ans))
    # extracts problem type from ANS i.e num_cmp, fun_cmp etc.
    answer_type_raw = re.findall(r'\((.*?)\(', str(question_ans))

    # determine answer type based on the raw answer from question
    for answer_type in answer_type_raw:
        if numerical_type in answer_type:
            question_type.append("number-input")
        elif functional_type in answer_type:
            question_type.append("functional")
        elif checkbox_type in answer_type:
            question_type.append("checkbox")
        elif text_type in answer_type:
            question_type.append("text")
        else:
            question_type.append("unknown")

    if not question_type:
        question_type.append("unknown")

    return {
        'answer_variable': answer_variable,
        'answer_type_raw': answer_type_raw,
        'question_type': question_type,
        'filename': filename}


def server(question_solution):
    # server variables
    server_imports = """
import random
import problem_bank_helpers as pbh
""".strip('\n')
    server_generate_names = "# TBD"
    server_generate_phrases = "# TBD"
    server_generate_random_var = "# N/A"
    if len(question_solution) > 0:
        server_generate_random_var = ''.join(f'# {solution.replace("$","").replace("**", "E").strip()}\n' for solution in question_solution)
    server_generate_dic = "# TBD"
    server_generate_answers = "# TBD"
    server_generate = f"""data2 = pbh.create_data2()\n# define or load names/items/objects from server files\n{server_generate_names}\n# store phrases etc\n{server_generate_phrases}\n# Randomize Variables\n{server_generate_random_var}\n# store the variables in the dictionary params\n{server_generate_dic}\n# define possible answers\n{server_generate_answers}\n# Update the data object with a new dict\ndata.update(data2)"""
    server_prepare = """pass
"""
    server_parse = """pass
"""
    server_grade = """pass
"""
    return {'imports': server_imports,
            'generate': server_generate,
            'prepare': server_prepare,
            'parse': server_parse,
            'grade': server_grade}

def yaml_dump(directory_info, metadata, question_format, image_dic, question_text, question_units, question_parts, question_solution, destination_file_path):
    # This solution is copied from this SO answer: https://stackoverflow.com/a/45004775/2217577
    yaml.SafeDumper.org_represent_str = yaml.SafeDumper.represent_str

    def repr_str(dumper, data):
        if '\n' in data:
            return dumper.represent_scalar(u'tag:yaml.org,2002:str', data, style='|')
        return dumper.org_represent_str(data)

    yaml.add_representer(str, repr_str, Dumper=yaml.SafeDumper)

    source = f"https://github.com/open-resources/webwork-open-problem-library/tree/master/{directory_info['file_dir']}"
    yaml_dict = {}

    # populate yaml_dic with values from args
    yaml_dict['title'] = metadata['title']
    yaml_dict['topic'] = metadata['topic']
    yaml_dict['author'] = metadata['author']
    #yaml_dict['date'] = metadata['date']
    yaml_dict['source'] = source
    yaml_dict['template_version'] = 1.3
    yaml_dict['attribution'] = 'standard'
    yaml_dict['partialCredit'] = 'true'
    yaml_dict['singleVariant'] = 'false'
    #yaml_dict['editor'] = metadata['editor']
    # yaml_dict['type'] = question_type['question_type']
    yaml_dict['outcomes'] = ['TBD']
    yaml_dict['difficulty'] = ['TBD']
    yaml_dict['randomization'] = ['TBD']
    yaml_dict['taxonomy'] = ['TBD']
    yaml_dict['span'] = ['TBD']
    yaml_dict['length'] = ['TBD']
    yaml_dict['tags'] = metadata['tags']
    yaml_dict['assets'] = image_dic['image_name'] if image_dic['image_name'] else ''
    for image in image_dic['image_name']:
        image_dir = directory_info['folder_dir'] + "/" + image
        image_path = Path(directory_info['folder_dir'] + "/" + image)
        if image_path.is_file():
            copy2(image_dir, destination_file_path)
    yaml_dict['server'] = server(question_solution)
    for part_number, part_type in zip(question_parts, question_format):
        if part_number + 1 == 0:
            part_number = str(part_number+2)
        else:
            part_number = str(part_number+1)
        yaml_dict['part' + part_number] = get_part_type(part_type)
    question_images = image_dic['image_line_md']
    Path(destination_file_path + directory_info['filename'] + ".md").write_text('---\n'
                                                                                + yaml.safe_dump(yaml_dict, sort_keys=False)
                                                                                + '---\n\n'
                                                                                + '# {{ params.vars.title }} \n\n'
                                                                                + ''.join(f'{image}\n' for image in question_images)
                                                                                + ''.join(f'\n{question}\n' for part, question in zip(question_parts, question_text) if (part == 0))
                                                                                + ''.join(f'\n## Part {part} \n{question} \n\n\n ### Answer Section\n' for part, question in zip(question_parts, question_text) if (part > 0))
                                                                                + str(question_units) + '\n\n'
                                                                                + '## pl-submission-panel \n\n\n'
                                                                                + '## pl-answer-panel \n\n\n'
                                                                                + '## Rubric \n\n\n'
                                                                                + '## Solution \n\n\n'
                                                                                + '## Comments \n\n\n')
                                                                                # + ''.join(f'{value}' for key, value in section.items())


def get_part_type(part_type):
    return {"type": part_type,
            "pl-customizations":
            {"weight": "1",
            "hide-answer-panel": "true"}
            }


def image_extract(question_content):
    image_src = "image("
    image_line = []
    image_alt_text = []
    image_name = []

    # iterate through each question part in the questions
    for question_part in question_content:
        # determine if the question contains an image
        if image_src in question_part:
            # extract image name and alt text using regex
            image_name = re.findall(' "(.+?)"', str(question_content).strip())
            image_alt_text = re.findall('="(.+?)"', str(question_content).strip())

    # bundle up the image name and alt text to create a .md image line
    for image_alt, image_filename in zip(image_alt_text, image_name):
        if image_alt:
            image_line.append('![' + image_alt.strip() + '](' + image_filename + ')')
        else:
            image_line.append('![](' + image_filename + ')')

    return {'image_name': image_name,
            'image_alt_text': image_alt_text,
            'image_line_md': image_line}


def problem_extract(question_body, image_alt_text):
    hint = ''
    question_units = ''
    question_raw = []
    question_split = ''
    part_headers = []
    question_part = []

    # split question into sections based on "$PAR"
    for question in question_body:
        question_split = question.split('$PAR\n')

    # for each section of the question
    for question_section in question_split:
        # if the section is not empty
        if len(question_section) > 0:
            # remote all the \n in the section
            section_clean = question_section.replace('\n', '')
            # find and assign hint (if it exits)
            if section_clean.endswith('</strong>') or section_clean.endswith('</b>'):
                hint = section_clean
            # if hint has not been assigned (no hint exists) OR section does NOT include hint
            if not hint or hint not in section_clean:
                subsection = help_problem_extract_ans_units(section_clean)
                subsection_text = subsection['section']
                question_units = subsection['final_ans_units']
                subsection_multi_part = help_problem_extract_ans_type(subsection_text)
                subsection_multi_part_ans_type = subsection_multi_part['ans_type']
                subsection_clean = subsection_multi_part['problem_clean']
                question_raw = help_problem_extract_append(subsection_clean, question_raw)
                # remove image_alt_text from question_raw and ensure there are no empty questions
                question_raw = [question for question in question_raw if question not in image_alt_text and question]
                question_part = append_part_counter(len(question_raw)-1, part_headers)

    return {'question_text': question_raw,
            'question_parts': question_part,
            'question_units': question_units}


def append_part_counter(part_counter, part_headers):
    if part_counter not in part_headers:
        part_headers.append(part_counter)
    return part_headers


def extract_problem_type(problem_subsection, filename):
    question_format_raw = re.findall("(ANS\(.+?\);)", str(problem_subsection))
    return determine_problem_type(question_format_raw, filename)


def help_problem_extract_ans_units(problem_subsection):
    final_ans_units = ''
    section_clean = ''
    if not problem_subsection.startswith("\\{ image") and not problem_subsection.endswith(") \\}"):
        # if section is the end i.e. ans_rule (determines the length of the answer)
        if problem_subsection.startswith("\\{ans_rule") and problem_subsection.endswith("\\)"):
            # extract the question units using regex
            final_ans_units = re.findall('textrm{(.+?)}', problem_subsection)
        if not problem_subsection.startswith("\\{ans_rule") and not problem_subsection.endswith("\\)"):
            section_clean = problem_subsection
    return {'section': section_clean,
            'final_ans_units': final_ans_units}


def help_problem_extract_ans_type(problem_subsection):
    ans_type = []
    problem_ans_type_removed = []
    if problem_subsection.startswith("END_TEXT"):
        ans_type = re.search(r"END_TEXT(.*)BEGIN_TEXT", str(problem_subsection))
        problem_ans_type_removed = re.sub(r"END_TEXT(.*)BEGIN_TEXT", "", str(problem_subsection))
    else:
        problem_ans_type_removed = problem_subsection

    return {'ans_type': ans_type,
            'problem_clean': problem_ans_type_removed}


def help_problem_extract_append(problem_subsection, final_dic):
    if len(problem_subsection) > 1:
        problem_stripped = problem_subsection.replace('\\', '').replace('textrm', '').replace('{', '').replace('}', '')\
            .replace('&middot;', '$\\cdot$').replace('END_TEXT', '').replace('BEGIN_TEXT', '').strip()
        if re.match(r'.\) ', problem_stripped):
            subsection_without_part_num = problem_stripped[3:]
            final_dic.append(subsection_without_part_num)
        else:
            final_dic.append(problem_stripped)

    return final_dic


def extract_problem_solution(problem_solution):
    question_solution = []

    for solution in problem_solution:
        split_solution = solution.split('\n')

        for solution_part in split_solution:
            if len(solution_part) > 0:
                # removes lines with showPartialCorrectAnswers and showHint
                if partial_answer_src not in solution_part and variables_start_src not in solution_part:
                    # removes lines with Context and TEXT(beginproblem());
                    if context_src not in solution_part and marcos_end_src not in solution_part:
                        question_solution.append(solution_part)

    return question_solution

# for loop runs based # of folders in src
for root, dirs, files in os.walk(root_path):
    # create dest file structure based on source directory
    for dir_path in src_dirs:
        Path(dir_path).mkdir(parents=True, exist_ok=True)
    # iterate through each file
    for file in files:
        if file.endswith('.pg'):
            source_files.append(os.path.join(root, file))


for source_filepath in source_files:
    try:
        dest_file_path = source_filepath[78:source_filepath.rfind('/')]
        filename = source_filepath[source_filepath.rfind('/')+1:-3]
        folder_dir = source_filepath[:source_filepath.rfind('/')]
        file_start_time = time.process_time()
        file_dir = source_filepath[source_filepath.find("Contrib"):]
        question_file = open(source_filepath, 'r')
        file_contents = question_file.read()

        file_contents_dic = split_file(file_contents)
        metadata_dic = metadata_extract(file_contents_dic['metadata'])
        dir_info = {
            'filename': filename,
            'file_dir': file_dir,
            'folder_dir': folder_dir,
            'root_dest_folder': root_dest_folder,
            'dest_file_path': dest_file_path
        }
        destination_file_path = root_dest_folder + dest_file_path + "/" + filename + "/"
        question_body = file_contents_dic['question_body']
        image_dic = image_extract(question_body)
        question_extract = problem_extract(question_body, image_dic['image_alt_text'])
        question_text = question_extract['question_text']
        question_parts = question_extract['question_parts']
        question_units = question_extract['question_units']
        question_formats = extract_problem_type(file_contents, dir_info['filename'])['question_type']
        question_solution = extract_problem_solution(file_contents_dic['question_solution'])
        Path(destination_file_path).mkdir(parents=True, exist_ok=True)
        yaml_dump(dir_info, metadata_dic, question_formats, image_dic, question_text,
                  question_units, question_parts, question_solution, destination_file_path)
        end_file_time = time.process_time()
        file_process_time = end_file_time - file_start_time
        counterString = '#' + str(counter + 1) + ' - [' + str(round(file_process_time, 5)) + '] '
        currentFile = root_dest_folder + dest_file_path + "/" + filename
        counter += 1
        print(counterString + currentFile)

    except Exception as e:
        print(e)
        pass

# ------------------------ STATS ------------------------ #
total_end_time = time.process_time()
process_time_seconds = total_end_time - total_start_time
print('total time:', round(process_time_seconds / 60, 2), 'minutes,', round(process_time_seconds, 2), 'seconds')
print('avg time per each file:', round(process_time_seconds / counter, 2), 'seconds [', counter, '] files')



# Author: Firas Moosvi and Graham Bovett
# Date: 2021-05-09
# This file contains many helper functions that will be used across the question bank project.

from docopt import docopt

# Imports
## Loading and Saving files & others
import uuid
import json
import pathlib
import sys
import numpy as np
import os
from collections import defaultdict
from shutil import copy2
import re
import codecs
import importlib.util
import problem_bank_helpers as pbh

## Parse Markdown
import markdown_it
import mdformat

## Dealing with YAML
import yaml

# Start of reading/parsing functions


def defdict_to_dict(defdict, finaldict):
    """Convert a defaultdict (nested) to a regular dictionary.
        - Answer copied from: https://stackoverflow.com/a/61133504/2217577
    Args:
        defdict (dict): defaultdict
        finaldict (dict): regular dictionary

    Returns:
        dict: Convert to regular dictionary
    """
    # pass in an empty dict for finaldict
    for k, v in defdict.items():
        if isinstance(v, defaultdict):
            # new level created and that is the new value
            finaldict[k] = defdict_to_dict(v, {})
        else:
            finaldict[k] = v
    return finaldict


def parse_body_part(pnum, md_text):
    """Parses markdown and returns a dictionary split by header

    Args:
        md_text (str): A string of markdown format

    Returns:
        parsed_md_text (dict): A dictionary split of the text
    """

    part = "part" + f"{pnum}"

    # Special dict to store stuff
    nested_dict = pbh.create_data2()

    # Create Markdown parser
    mdit = markdown_it.MarkdownIt()
    env = {}
    tokens = mdit.parse(md_text, env)

    # Get Level 2 headers and make sure there's only one!
    level2_headers = [i for i, j in enumerate(tokens) if j.tag == "h2" if j.nesting == 1]
    assert (
        len(level2_headers) == 1
    ), "There is a problem in the question, there seem to be multiple level two headers in a body part, or there is a weird edge-case in the parse_body_part() function"

    assert (
        len(tokens[level2_headers[0] + 1].content) < 20
    ), "There is an (arbitrary/opinionated) restriction on the length of 20 chars for a a 'part' title."

    nested_dict[part]["title"] = tokens[level2_headers[0] + 1].content

    # Store the content of the level 2 header
    try:
        content = codecs.unicode_escape_decode(
            mdformat.renderer.MDRenderer().render(
                tokens[3 : get_next_headerloc(3, tokens, 3)], mdit.options, env
            )  # Note the 3 is there to exclude header start,header content,header end tokens
        )[0]
        nested_dict[part]["content"] = content
    except IndexError:
        print("It looks like there is an empty section of header level 2 in your md file.")
        raise

    # Get all Level 3 headers
    level3_headers = [i for i, j in enumerate(tokens) if j.tag == "h3" if j.nesting == 1]

    for hd in level3_headers:

        header = tokens[hd + 1].content
        assert (
            len(header) < 20
        ), "There is an (arbitrary/opinionated) restriction on the length of 20 chars for a a 'sub-part' title."
        if "Answer Section" in header:
            header = "answer"
        try:
            content = codecs.unicode_escape_decode(
                mdformat.renderer.MDRenderer().render(
                    tokens[hd + 3 : get_next_headerloc(hd + 3, tokens, 3)], mdit.options, env
                )
            )[
                0
            ]  # Note the +3 is there to exclude header start,header content,header end tokens
            nested_dict[part][header] = content
        except IndexError:
            print("It looks like there is an empty section of header level 3 in your md file.")
            # TODO: in the future, suggest ignoring empty sections instead of throwing an error
            raise

    return defdict_to_dict(nested_dict, {})


def get_next_headerloc(start, tokens, header_level):
    """Some annoying code that takes in a token to start looking from, and returns the token at the start of the next header level.

    Args:
        start (int): Token # to start looking from
        tokens (token): Parsed MD file
        header_level (int): "Next" Header level to look for.

    Returns:
        int: Token # to "end" looking at
    """

    # no header found, set close to end
    close = len(tokens)

    for i, j in enumerate(tokens[start:]):

        if j.tag == f"h{header_level}" and j.nesting == 1:
            # next header found
            close = i + start
            break
    return close


# def split_body_parts(num_parts, body_parts):
#     """Parses individual question parts and splits out titles, and content

#     Args:
#         num_parts (int): An integer corresponding to the number of question parts (from `read_md_problem()`).
#         body_parts (dict): A dictionary from `read_md_problem()`.

#     Returns:
#         body_parts_dict (dict): returns a nested dictionary with title,content,answer keys .
#     """
#     mdit = markdown_it.MarkdownIt()
#     env = {}
#     nested_dict = lambda: defaultdict(nested_dict)

#     parts_dict = nested_dict()

#     for pnum in range(1, num_parts + 1):

#         part = "part" + f"{pnum}"
#         # Set up tokens by parsing the md file
#         tokens = mdit.parse(body_parts[part], env)

#         ptt = [i for i, j in enumerate(tokens) if j.tag == "h2"]
#         parts_dict[part]["title"] = (
#             mdformat.renderer.MDRenderer()
#             .render(tokens[ptt[0] + 1 : ptt[1]], mdit.options, env)
#             .strip("\n")
#         )

#         # Get the "### Answer section from the parts_dict"
#         pa = [i
#                 for i, j in enumerate(tokens)
#                 if j.tag == "h3"
#                 if "pl-submission-panel" not in j.content
#                 if "pl-answer-panel" not in j.content
#             ]

#         try:
#             parts_dict[part]["answer"]["title"] = codecs.unicode_escape_decode(
#                 mdformat.renderer.MDRenderer().render(
#                     tokens[pa[0] + 1 : pa[1]], mdit.options, env
#                 )
#             )[0]
#         except IndexError:
#             print(
#                 "Check the heading levels, is there one that doesn't belong? Or is the heading level incorrect? For e.g., it should be ### Answer Section (this is not necessarily where the issue is)."
#             )
#             raise

#         parts_dict[part]["content"] = codecs.unicode_escape_decode(
#             mdformat.renderer.MDRenderer().render(
#                 tokens[ptt[1] + 1 : pa[0]], mdit.options, env
#             )
#         )[0]
#         parts_dict[part]["answer"]["content"] = codecs.unicode_escape_decode(
#             mdformat.renderer.MDRenderer().render(
#                 tokens[pa[1] + 1 :], mdit.options, env
#             )
#         )[0]

#         # Get the ### pl-submission-panel and ### pl-answer-panel
#         p_extra = [i
#                 for i, j in enumerate(tokens)
#                 if j.tag == "h3"
#                 if "pl-submission-panel" in j.content
#                 if "pl-answer-panel" in j.content
#             ]

#         if p_extra:
#             try:
#                 parts_dict[part]["pl-submission-panel"] = codecs.unicode_escape_decode(
#                     mdformat.renderer.MDRenderer().render(
#                         tokens[p_extra[0] + 1 : p_extra[1]], mdit.options, env
#                     )
#                 )[0]
#             except IndexError:
#                 print(
#                     "Check the heading levels, is there one that doesn't belong? Or is the heading level incorrect? For e.g., it should be ### Answer Section (this is not necessarily where the issue is)."
#                 )
#                 raise


#         # for key in body_parts.keys():
#         #     print(f'outside: {key}')
#         #     if key in ['pl-submission-panel','pl-answer-panel']:
#         #         # Set up tokens by parsing the md file
#         #         tokens = mdit.parse(body_parts[key], env)

#         #         print(key)

#         #         ptt = [i for i,j in enumerate(tokens) if j.tag=='h3']
#         #         try:
#         #             parts_dict[part][key] = codecs.unicode_escape_decode(MDRenderer().render(tokens[ptt[-1]+1:], mdit.options, env))[0]
#         #         except IndexError:
#         #             print("It's possible you have '### pl-submission-panel' or '### pl-answer-panel' with the wrong heading level - H2 instead of the required H3.")

#         # Remove parts from body_parts
#         body_parts.pop(part)

#     return defdict_to_dict(parts_dict, {})


def read_md_problem(filepath):
    """Reads a MystMarkdown problem file and returns a dictionary of the header and body

    Args:
        filepath (str): Path of file to read.

    Returns:
        dict: In this dictionary there are three keys containing useful portions of the parsed md file:
            - `header` - Header of the problem file (nested dictionary).
            - `body_parts` - Body text of the problem file (nested dictionary).
            - `num_parts` - Number of parts in the problem (integer).
            - `body_parts_split` - Dictionary with each part split into individual components.
    """

    mdtext = pathlib.Path(filepath).read_text(encoding="utf8")

    # Deal with YAML header
    header_text = mdtext.rsplit("---\n")[1]
    header = yaml.safe_load("---\n" + header_text)

    # Deal with Markdown Body
    body = mdtext.rsplit("---\n")[2]

    # Set up the markdown parser
    # to be honest, not fully sure what's going on here, see this issue: https://github.com/executablebooks/markdown-it-py/issues/164

    mdit = markdown_it.MarkdownIt()
    env = {}

    # Set up tokens by parsing the md file
    tokens = mdit.parse(body, env)

    blocks = {}

    block_count = 0

    num_titles = 0

    ###
    for x, t in enumerate(tokens):

        if t.tag == "h1" and t.nesting == 1:  # title
            # oh boy. this is going to break and it will be your fault firas.
            blocks["title"] = [x, x + 3]
            num_titles += 1

        elif t.tag == "h2" and t.nesting == 1:
            block_count += 1

            if block_count == 1:
                blocks["block{0}".format(block_count)] = [
                    x,
                ]
            else:
                blocks["block{0}".format(block_count - 1)].append(x)
                blocks["block{0}".format(block_count)] = [
                    x,
                ]
    ###
    # Add -1 to the end of the last block
    blocks[f"block{block_count}"].append(len(tokens))

    # Assert statements (turn into tests!)
    assert num_titles == 1, "I see {0} Level 1 Headers (#) in this file, there should only be one!".format(num_titles)
    assert block_count >= 1, "I see {0} Level 2 Headers (##) in this file, there should be at least 1".format(
        block_count - 1
    )

    # Add the end of the title block; # small hack
    # blocks['title'].append(blocks['block1'][0])

    # Get the preamble before the parts start
    blocks["preamble"] = [blocks["title"][1], blocks["block1"][0]]

    ## Process the blocks into markdown

    body_parts = {}
    parts_dict = {}

    part_counter = 0

    for k, v in blocks.items():

        rendered_part = codecs.unicode_escape_decode(
            mdformat.renderer.MDRenderer().render(tokens[v[0] : v[1]], mdit.options, env)
        )[0]

        if k == "title":
            body_parts["title"] = rendered_part

        elif k == "preamble":
            body_parts["preamble"] = rendered_part

        elif "Rubric" in rendered_part:
            body_parts["Rubric"] = rendered_part

        elif "Solution" in rendered_part:
            body_parts["Solution"] = rendered_part

        elif "Comments" in rendered_part:
            body_parts["Comments"] = rendered_part

        elif "Useful Info" in rendered_part:
            body_parts["Useful_info"] = rendered_part

        else:
            part_counter += 1
            body_parts[f"part{part_counter}"] = rendered_part

            parts_dict.update(parse_body_part(part_counter, rendered_part))

    return_dict = {
        "header": header,
        "body_parts": body_parts,
        "num_parts": part_counter,
        "body_parts_split": parts_dict,
    }
    return defdict_to_dict(return_dict, {})


def dict_to_md(
    md_dict,
    remove_keys=[
        None,
    ],
):
    """Takes a nested dictionary (e.g. output of read_md_problem()) and returns a multi-line string  that can be written to a file (after removing specified keys).
    Args:
        md_dict (dict): A nested dictionary, for e.g. the output of `read_md_problem()`
        remove_keys (list, optional): Any keys to remove from the dictionary, for instance solutions. Defaults to [None,].

    Returns:
        str: A multi-line string that can be written to a file.
    """

    md_string = ""

    md_dict = defdict_to_dict(md_dict, {})

    # Question Title and Preamble
    md_string += md_dict.pop("title", None)
    md_string += md_dict.pop("preamble", None)

    #TODO: Refactor this to use the elegant solution provided here: https://stackoverflow.com/a/49723101/2217577

    for k, v in md_dict.items():
        if k in remove_keys:
            continue
        else:
            md_string += "\n" + md_dict[k]

    return md_string


## Functions from md-to-pl


def write_info_json(output_path, parsed_question):
    """
    Args:
        output_path (Path): [description]
        parsed_question (dict]): [description]
    """

    # Deal with optional tags in info.json
    optional = ""

    if parsed_question["header"].get("gradingMethod"):
        optional += """ "gradingMethod": parsed_question['header']['gradingMethod'],\n"""
    elif parsed_question["header"].get("partialCredit"):
        optional += """ "partialCredit": parsed_question['header']['partialCredit'],\n"""
    elif parsed_question["header"].get("externalGradingOptions"):
        optional += """ "externalGradingOptions": parsed_question['header']['externalGradingOptions'],\n"""
    elif parsed_question["header"].get("dependencies"):
        optional += """ "dependencies": parsed_question['header']['dependencies'],\n"""
    elif parsed_question["header"].get("singleVariant"):
        optional += """ "singleVariant": parsed_question['header']['singleVariant'],\n"""
    elif parsed_question["header"].get("showCorrectAnswer"):
        optional += """ "showCorrectAnswer": parsed_question['header']['showCorrectAnswer'],\n"""

    # Add tags based on part type
    q_types = []

    for pnum in range(1, parsed_question["num_parts"] + 1):
        part = "part" + f"{pnum}"
        q_types.append(parsed_question["header"][part]["type"])

    auto_tags = []
    if len(q_types) > 1:
        auto_tags.append("multi_part")
    auto_tags.extend(list(set(q_types)))

    auto_tags.extend(parsed_question["header"]["tags"])
    auto_tags = [v for v in auto_tags if v != "unknown"]

    # End add tags

    pathlib.Path(output_path / "info.json").write_text(
        """{
            "uuid": \""""
        + str(uuid.uuid3(uuid.NAMESPACE_DNS, str(output_path)))
        + """\",
            "title": \""""
        + parsed_question["header"]["title"]
        + """",
            "topic": \""""
        + parsed_question["header"]["topic"]
        + """",
            "tags":  """
        + json.dumps(auto_tags)
        + """,
            "type": "v3"
        }""",
        encoding="utf8",
    )


def assemble_server_py(parsed_question, location):
    """Assembles a string version of the server.py file from the YAML header of the md file.

    Args:
        parsed_question (_type_): dictionary that is created upon reading of the md problem.
        location (string): 'local' or 'prairielearn' ; the import statements are different depending on if it's local or on a PL server.
    """

    server_dict = parsed_question["header"]["server"]

    if location == "local":
        # This is needed to run this locally compared to when it gets run on a PL server
        server_dict["imports"] = parsed_question["header"]["server"]["imports"].replace(
            "import prairielearn as pl",
            "import problem_bank_scripts.prairielearn as pl",
        )

    server_py = f""""""

    server_py += server_dict.get("imports", "") + "\n"

    try:
        for function, code in server_dict.items():
            indented_code = code.replace("\n", "\n    ")
            if code:
                server_py += f"def {function}(data):\n    {indented_code}\n"
    except:
        raise

    return server_py


def write_server_py(output_path, parsed_question):
    """Writes the server.py file to disk
    Args:
        output_path ([type]): [description]
        parsed_question ([type]): [description]
    """

    output_path = pathlib.Path(output_path)

    server_file = assemble_server_py(parsed_question, "prairielearn")

    # Deal with path differences when using PL
    server_file = server_file.replace('read_csv("', 'read_csv(data["options"]["client_files_course_path"]+"/')

    # Write server.py
    (output_path / "server.py").write_text(server_file, encoding="utf8")


def process_multiple_choice(part_name, parsed_question, data_dict):
    """Processes markdown format multiple-choice questions and returns PL HTML
    Args:
        output_path (Path): [description]
        parsed_question (dict): [description]
        data_dict (dict)

    Returns:
        str: Multiple choice question is returned as a string with PL-compliant syntax.
    """

    html = f"""<pl-question-panel>\n<markdown>{parsed_question['body_parts_split'][part_name]['content']}</markdown>\n</pl-question-panel>\n\n"""

    pl_customizations = " ".join(
        [f'{k} = "{v}"' for k, v in parsed_question["header"][part_name]["pl-customizations"].items()]
    )  # PL-customizations
    html += f"""<pl-multiple-choice answers-name="{part_name}_ans" {pl_customizations} >\n"""

    ###### LOOKHERE
    if (data_dict["params"][f"vars"]["units"]) and (
        "units" in parsed_question["body_parts_split"][part_name]["answer"]
    ):
        units = f"|@ params.vars.units @|"
    else:
        units = ""

    ## Note: `|@`` gets converted into `{{` and `@|`` gets converted to `}}` by `replace_tags()`
    for a in data_dict["params"][f"{part_name}"].keys():

        if "ans" in a:
            if data_dict["params"][f"{part_name}"][f"{a}"]["feedback"]:
                feedback = f"|@ params.{part_name}.{a}.feedback @|"
            else:
                feedback = f"Feedback for this choice is not available yet."

            correctness = f"|@ params.{part_name}.{a}.correct @|"
            value = f"|@ params.{part_name}.{a}.value @|"

            html += f"\t<pl-answer correct= {correctness} feedback = '{feedback}' > {value} {units} </pl-answer>\n"

    html += "</pl-multiple-choice>\n"

    return replace_tags(html)


def process_dropdown(part_name, parsed_question, data_dict):
    """Processes markdown format dropdown questions and returns PL HTML

    Args:
        part_name (string): Name of the question part being processed (e.g., part1, part2, etc...)
        parsed_question (dict): Dictionary of the MD-parsed question (output of `read_md_problem`)
        data_dict (dict): Dictionary of the `data` dict created after running server.py using `exec()`

    Returns:
        html: A string of HTML that is part of the final PL question.html file.
    """
    html = process_multiple_choice(part_name, parsed_question, data_dict).replace("-multiple-choice", "-dropdown")
    return html


def process_number_input(part_name, parsed_question, data_dict):
    """Processes markdown format number-input questions and returns PL HTML

    Args:
        part_name (string): Name of the question part being processed (e.g., part1, part2, etc...)
        parsed_question (dict): Dictionary of the MD-parsed question (output of `read_md_problem`)
        data_dict (dict): Dictionary of the `data` dict created after running server.py using `exec()`

    Returns:
        html: A string of HTML that is part of the final PL question.html file.
    """

    html = f"""<pl-question-panel>\n\t<markdown>{parsed_question['body_parts_split'][part_name]['content']}\t</markdown>\n</pl-question-panel>\n\n"""

    pl_customizations = " ".join(
        [f'{k} = "{v}"' for k, v in parsed_question["header"][part_name]["pl-customizations"].items()]
    )  # PL-customizations
    html += f"""<pl-number-input answers-name="{part_name}_ans" {pl_customizations} ></pl-number-input>\n"""

    return replace_tags(html)


def process_checkbox(part_name, parsed_question, data_dict):
    """Processes markdown format checkbox (select all that apply) questions and returns PL HTML

    Args:
        part_name (string): Name of the question part being processed (e.g., part1, part2, etc...)
        parsed_question (dict): Dictionary of the MD-parsed question (output of `read_md_problem`)
        data_dict (dict): Dictionary of the `data` dict created after running server.py using `exec()`

    Returns:
        html: A string of HTML that is part of the final PL question.html file.
    """
    # start with the MCQ version and then...change things for checkbox questions
    html = process_multiple_choice(part_name, parsed_question, data_dict).replace("-multiple-choice", "-checkbox")
    return html


def process_symbolic_input(part_name, parsed_question, data_dict):
    """Processes markdown format symbolic questions and returns PL HTML

    Args:
        part_name (string): Name of the question part being processed (e.g., part1, part2, etc...)
        parsed_question (dict): Dictionary of the MD-parsed question (output of `read_md_problem`)
        data_dict (dict): Dictionary of the `data` dict created after running server.py using `exec()`

    Returns:
        html: A string of HTML that is part of the final PL question.html file.
    """

    html = f"""<pl-question-panel>\n\t<markdown>{parsed_question['body_parts_split'][part_name]['content']}\t</markdown>\n</pl-question-panel>\n\n"""

    pl_customizations = " ".join(
        [f'{k} = "{v}"' for k, v in parsed_question["header"][part_name]["pl-customizations"].items()]
    )  # PL-customizations
    html += f"""<pl-symbolic-input answers-name="{part_name}_ans" {pl_customizations} ></pl-symbolic-input>\n"""

    return replace_tags(html).replace("\\\\", "\\")


def process_longtext(part_name, parsed_question, data_dict):
    """Processes markdown format of long-text questions and returns PL HTML
    Args:
        part_name (string): Name of the question part being processed (e.g., part1, part2, etc...)
        parsed_question (dict): Dictionary of the MD-parsed question (output of `read_md_problem`)
        data_dict (dict): Dictionary of the `data` dict created after running server.py using `exec()`

    Returns:
        html: A string of HTML that is part of the final PL question.html file.
    """
    pl_customizations = " ".join(
        [f'{k} = "{v}"' for k, v in parsed_question["header"][part_name]["pl-customizations"].items()]
    )  # PL-customizations

    html = f"""<pl-question-panel>\n<markdown>{parsed_question['body_parts_split'][part_name]['content']}</markdown>\n</pl-question-panel>\n\n"""

    html += f"""<pl-rich-text-editor { pl_customizations } > </pl-rich-text-editor>"""

    return replace_tags(html)


def process_file_upload(part_name, parsed_question, data_dict):
    """Processes markdown format of file-upload questions and returns PL HTML
    Args:
        part_name (string): Name of the question part being processed (e.g., part1, part2, etc...)
        parsed_question (dict): Dictionary of the MD-parsed question (output of `read_md_problem`)
        data_dict (dict): Dictionary of the `data` dict created after running server.py using `exec()`

    Returns:
        html: A string of HTML that is part of the final PL question.html file.
    """
    pl_customizations = " ".join(
        [f'{k} = "{v}"' for k, v in parsed_question["header"][part_name]["pl-customizations"].items()]
    )  # PL-customizations

    html = f"""<pl-question-panel>\n<markdown>{parsed_question['body_parts_split'][part_name]['content']}</markdown>\n</pl-question-panel>\n\n"""

    html += f"""<pl-file-upload { pl_customizations } > </pl-file-upload>"""

    html += f"""<pl-submission-panel>\n\t<pl-file-preview></pl-file-preview>\n\t<pl-external-grader-results></pl-external-grader-results>"""

    # TODO: remove this! because automatic feedback will be added
    html += f"""\n\t|@ #feedback.manual @| \n\t<p>Feedback from course staff:</p>\n\t<markdown>|@|@ feedback.manual @|@|</markdown>\n\t|@ /feedback.manual @|\n</pl-submission-panel>"""

    # TODO: Add better support for what students see when they upload a file where many are possible. Currently: Error: The following required files were missing: *.jpg, *.pdf, foo.py, bar.c, filename space.txt
    # TODO: Add support for wildcard *.png
    # TODO: Add better message telling students the question needs to be manually graded.

    return replace_tags(html)


def replace_tags(string):
    """Takes in a string with tags: |@ and @| and returns {{ and }} respectively. This is because Python strings can't have double curly braces.

    Args:
        string (str): String to be processed, can be multi-line.

    Returns:
        string (str): returns string with tags replaced with curly braces.
    """
    return string.replace("|@|@", "{{{").replace("@|@|", "}}}").replace("|@", "{{").replace("@|", "}}")


def remove_correct_answers(data2_dict):
    """Magical recursive function that removes particular keys from a nested dictionary: https://stackoverflow.com/a/29652561/2217577

    Args:
        data2_dict (dict): Dictionary (nested) from which to remove key:value

    Returns:
        data2_dict (dict): Dictionary with the offending keys removed
    """

    # This was adapted from this SO: https://stackoverflow.com/a/29652561/2217577
    def gen_dict_extract(key_to_remove, dict_object):
        if hasattr(dict_object, "items"):
            for k, v in list(dict_object.items()):
                if key_to_remove in k:
                    dict_object.pop(k, None)
                if isinstance(v, dict):
                    for result in gen_dict_extract(key_to_remove, v):
                        yield result
                elif isinstance(v, list):
                    for d in v:
                        for result in gen_dict_extract(key_to_remove, d):
                            yield result

    list(gen_dict_extract("correct", data2_dict))

    return data2_dict


def process_attribution(attribution):
    """Takes in a string and returns the HTML for the attribution

    Args:
        attribution (string): One of a set of pre-defined values corresponding to a particular attribution.

    Returns:
        string (str): returns the html of the attribution
    """

    possible_attributions = {
        "openstax-physics-vol1": "Problem is from the [OpenStax University Physics Volume 1](https://openstax.org/details/books/university-physics-volume-1) textbook, licensed under the [CC-BY 4.0 license](https://creativecommons.org/licenses/by/4.0/).<br>![Image representing the Creative Commons 4.0 BY license.](https://raw.githubusercontent.com/firasm/bits/master/by.png)",
        "openstax-physics-vol2": "Problem is from the [OpenStax University Physics Volume 2](https://openstax.org/details/books/university-physics-volume-2) textbook, licensed under the [CC-BY 4.0 license](https://creativecommons.org/licenses/by/4.0/).<br>![Image representing the Creative Commons 4.0 BY license.](https://raw.githubusercontent.com/firasm/bits/master/by.png)",
        # "ubc-mech2"
        "standard": "Problem is licensed under the [CC-BY-NC-SA 4.0 license](https://creativecommons.org/licenses/by-nc-sa/4.0/).<br> ![The Creative Commons 4.0 license requiring attribution-BY, non-commercial-NC, and share-alike-SA license.](https://raw.githubusercontent.com/firasm/bits/master/by-nc-sa.png)",
    }

    try:
        attribution_text = possible_attributions[attribution]

    except KeyError:
        print(
            f"`Attribution` value of {attribution} is not recognized. Currently, the only possible values are: {possible_attributions.keys()}. You need to update your md file and fix the `attribution` in the header"
        )
        raise

    return attribution_text


def process_question_md(source_filepath, output_path=None, instructor=False):

    try:
        pathlib.Path(source_filepath)
    except:
        print(f"{source_filepath} - File does not exist.")
        raise

    if output_path is None:
        if instructor:
            path_replace = "output/instructor"
        else:
            path_replace = "output/public"

        if "source" in source_filepath:
            output_path = pathlib.Path(source_filepath.replace("source", path_replace))
        else:
            raise NotImplementedError("Check the source filepath; it does not have 'source' in it!! ")
    else:
        ## TODO: Make this a bit more robust, perhaps by switching encodings!?
        output_path = pathlib.Path(output_path)

    # deal with multi-line strings in YAML Dump
    ## Code copied from here: https://stackoverflow.com/a/33300001/2217577

    def str_presenter(dumper, data2):
        if len(data2.splitlines()) > 1:  # check for multiline string
            # data2 = re.sub('\\n[\s].*\\n','\n\n',data2) # THIS IS WRONG!!!
            data2 = re.sub("\\n\s+\\n", "\n\n", data2)  # # Try \s{3,} for three or more spaces
            return dumper.represent_scalar("tag:yaml.org,2002:str", data2, style="|")
        return dumper.represent_scalar("tag:yaml.org,2002:str", data2)

    yaml.add_representer(str, str_presenter)

    parsed_q = read_md_problem(source_filepath)

    header = parsed_q["header"]
    body_parts = parsed_q["body_parts"]

    #################################################################################
    # Run the python code; this improved way was suggested by Phil Austin of UBC EOAS

    server_py = assemble_server_py(parsed_q, "local")

    spec = importlib.util.spec_from_loader("server", loader=None)
    server = importlib.util.module_from_spec(spec)
    exec(server_py, server.__dict__)

    data2 = pbh.create_data2()
    server.generate(data2)
    #################################################################################

    # Remove the solutions from the server section
    if instructor is False:
        # Remove python solution from the public view
        header.pop("server", None)

        # Remove correct answers from the data2 dict
        data2_sanitized = defdict_to_dict(data2, {})
        data2_sanitized = remove_correct_answers(data2_sanitized)

        # Update the YAML header to add substitutions
        header.update({"substitutions": defdict_to_dict(data2_sanitized, {})})

        # Update the YAML header to add substitutions, unsort it, and process for file
        header_yml = yaml.dump(header, sort_keys=False, allow_unicode=True)

        # Write the YAML to a file
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_text(
            "---\n"
            + header_yml
            + "---\n"
            + dict_to_md(
                body_parts,
                remove_keys=[
                    "Rubric",
                    "Solution",
                    "Comments",
                    "pl-submission-panel", #FIXME: This will not remove level 3 headings because it's all a string!
                    "pl-answer-panel",     #FIXME: This will not remove level 3 headings because it's all a string!
                ],
            )
            + "\n## Attribution\n\n"
            + process_attribution(header.get("attribution")),
            encoding="utf8",
        )

    else:
        # Update the YAML header to add substitutions
        header.update({"substitutions": defdict_to_dict(data2, {})})

        # return {'header':header,
        #         'body_parts':body_parts,
        #         'output_path':output_path}

        # Update the YAML header to add substitutions, unsort it, and process for file
        header_yml = yaml.dump(header, sort_keys=False, allow_unicode=True)

        # Write the YAML to a file
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_text(
            "---\n"
            + header_yml
            + "\n---\n"
            + dict_to_md(
                body_parts,
            )
            + "\n## Attribution\n\n"
            + process_attribution(header.get("attribution")),
            encoding="utf8",
        )

    # Move image assets
    files_to_copy = header.get("assets")
    if files_to_copy:
        [copy2(pathlib.Path(source_filepath).parent / fl, output_path.parent) for fl in files_to_copy]


def process_question_pl(source_filepath, output_path=None):

    try:
        pathlib.Path(source_filepath)
    except:
        print(f"{source_filepath} - File does not exist.")
        raise

    path_replace = "output/prairielearn"

    if output_path is None:
        if "source" in source_filepath:
            output_path = pathlib.Path(source_filepath.replace("source", path_replace)).parent
        else:
            raise NotImplementedError("Check the source filepath; it does not have 'source' in it!! ")
    else:
        ## TODO: It's annoying that here output_path.parent is used, but for md problems, it's just output_path
        output_path = pathlib.Path(output_path).parent

    # Parse the MD file
    parsed_q = read_md_problem(source_filepath)

    # Create output dir if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    #################################################################################
    # Run the python code; this improved way was suggested by Phil Austin of UBC EOAS

    server_py = assemble_server_py(parsed_q, "local")

    spec = importlib.util.spec_from_loader("server", loader=None)
    server = importlib.util.module_from_spec(spec)
    exec(server_py, server.__dict__)

    data2 = pbh.create_data2()

    server.generate(data2)
    #################################################################################

    # Write info.json file
    write_info_json(output_path, parsed_q)

    # Question Preamble
    preamble = parsed_q["body_parts"].get("preamble", None)
    #TODO: Remove Debugging print statement
    #print(f"premable: {preamble}")
    
    if preamble:
        question_html = f"<pl-question-panel>\n<markdown>\n{ preamble }\n</markdown>\n</pl-question-panel>\n\n"
    else:
        question_html = f""

    # Useful info panel
    useful_info = parsed_q["body_parts"].get("Useful_info", None)

    #TODO: When PrairieLearn updates to BootStrap5, update this box as described here: https://github.com/open-resources/problem_bank_scripts/issues/30#issuecomment-1177101211
    if useful_info:
        question_html += f"""<p>
   <button class="btn btn-primary" type="button" data-toggle="collapse" data-target="#collapseExample" aria-expanded="false" aria-controls="collapseExample">
   <i class="fa fa-info-circle" aria-hidden="true"></i> Helpful Information
   </button>
</p>
<div class="collapse" id="collapseExample">
   <div class="card card-body">
      <markdown>
{parsed_q['body_parts']['Useful_info']}
      </markdown>
   </div>
</div>"""

    # Single and Multi-part question construction

    for pnum in range(1, parsed_q["num_parts"] + 1):
        part = "part" + f"{pnum}"
        q_type = parsed_q["header"][part]["type"]

        question_html += f"\n<!-- ######## Start of Part {pnum} ######## -->\n\n"

        if parsed_q["num_parts"] > 1:
            question_html += f"""<div class="card my-2">
<div class="card-header">{parsed_q['body_parts_split'][part]['title']}</div>\n
<div class="card-body">\n\n"""

        if "multiple-choice" in q_type:
            question_html += f"{process_multiple_choice(part,parsed_q,data2)}"
        elif "number-input" in q_type:
            question_html += f"{process_number_input(part,parsed_q,data2)}"
        elif "checkbox" in q_type:
            question_html += process_checkbox(part, parsed_q, data2)
        elif "symbolic-input" in q_type:
            question_html += process_symbolic_input(part, parsed_q, data2)
        elif "dropdown" in q_type:
            question_html += process_dropdown(part, parsed_q, data2)
        elif "longtext" in q_type:
            question_html += process_longtext(part, parsed_q, data2)
        elif "file-upload" in q_type:
            question_html += process_file_upload(part, parsed_q, data2)
        else:
            raise NotImplementedError(f"This question type ({q_type}) is not yet implemented.")

        if parsed_q["num_parts"] > 1:
            question_html += "</div>\n</div>\n"

        # Add pl-submission-panel and pl-answer-panel (if they exist)
        subm_panel = parsed_q["body_parts_split"][part].get("pl-submission-panel", None)
        q_panel = parsed_q["body_parts_split"][part].get("pl-answer-panel", None)

        if subm_panel:
            question_html += f"\n<pl-submission-panel>{ subm_panel } </pl-submission-panel>\n"
        if q_panel:
            question_html += f"\n<pl-answer-panel>{ q_panel } </pl-answer-panel>\n"

        # TODO: Add support for other panels here as well !

        question_html += f"\n<!-- ######## End of Part {pnum} ######## -->\n"

    # Add Attribution
    question_html += f"\n<pl-question-panel>\n<markdown>\n---\n{process_attribution(parsed_q['header'].get('attribution'))}\n</markdown>\n</pl-question-panel>\n"

    # Fix Latex underscore bug (_ being replaced with \_)
    question_html = question_html.replace("\\_", "_")

    # Final pre-processing
    question_html = pl_image_path(question_html)

    # Write question.html file
    (output_path / "question.html").write_text(question_html, encoding="utf8")

    ### TODO solve the issue with the latex escape sequences, this is a workaround
    # with open((output_path / "question.html"), "w") as qfile:
    #     print(f"{question_html}", file=qfile)

    # Write server.py file
    write_server_py(output_path, parsed_q)

    # Move image assets
    files_to_copy = parsed_q["header"].get("assets")
    if files_to_copy:
        pl_path = output_path / "clientFilesQuestion"
        pl_path.mkdir(parents=True, exist_ok=True)
        [copy2(pathlib.Path(source_filepath).parent / fl, pl_path / fl) for fl in files_to_copy]


def pl_image_path(html):

    """Adds `{{options.client_files_question_url}}` directory before the path automatically"""

    # TODO: Figure out the regex to make this into a single expression, maybe with |
    # If image files are included as markdown format, add {{options.client_files_question_url}}
    res = re.subn(r"\(((?!http).*\.png)\)", "({{options.client_files_question_url}}/\\1)", html)
    res = re.subn(r"\(((?!http).*\.gif)\)", "({{options.client_files_question_url}}/\\1)", html)
    res = re.subn(r"\(((?!http).*\.jpeg)\)", "({{options.client_files_question_url}}/\\1)", html)
    res = re.subn(r"\(((?!http).*\.jpg)\)", "({{options.client_files_question_url}}/\\1)", html)

    # If image files are included as html format, add {{options.client_files_question_url}}
    res = re.subn(
        r"src[\s,=]*\"(?!http)(.*\.png)",
        'src="{{options.client_files_question_url}}/\\1',
        res[0],
    )  # works
    res = re.subn(
        r"src[\s,=]*\"(?!http)(.*\.gif)",
        'src="{{options.client_files_question_url}}/\\1',
        res[0],
    )  # works
    res = re.subn(
        r"src[\s,=]*\"(?!http)(.*\.jpeg)",
        'src="{{options.client_files_question_url}}/\\1',
        res[0],
    )  # works
    res = re.subn(
        r"src[\s,=]*\"(?!http)(.*\.jpg)",
        'src="{{options.client_files_question_url}}/\\1',
        res[0],
    )  # works

    return res[0]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['problem_bank_scripts']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'docopt>=0.6.2,<0.7.0',
 'markdown-it-py>=2.1.0,<3.0.0',
 'mdformat>=0.7.14,<0.8.0',
 'numpy>=1.22,<2.0',
 'pandas>=1.2.4,<2.0.0',
 'problem-bank-helpers>=0.1.12,<0.2.0',
 'sympy>=1.8,<2.0']

setup_kwargs = {
    'name': 'problem-bank-scripts',
    'version': '0.4.5',
    'description': 'A package with useful functions to convert between different problem bank formats.',
    'long_description': '# Problem Bank Scripts \n\n[![Python](https://img.shields.io/badge/python-3.9-blue)]()\n[![codecov](https://codecov.io/gh/open-resources/problem_bank_scripts/branch/main/graph/badge.svg)](https://codecov.io/gh/open-resources/problem_bank_scripts)\n[![Documentation Status](https://readthedocs.org/projects/problem_bank_scripts/badge/?version=latest)](https://problem_bank_scripts.readthedocs.io/en/latest/?badge=latest)\n\n\n## Installation\n\n```bash\n$ pip install -i https://test.pypi.org/simple/ problem_bank_scripts\n```\n\n## Features\n\n- TODO\n\n## Dependencies\n\n- TODO\n\n## Usage\n\n- TODO\n\n## Documentation\n\nThe official documentation is hosted on Read the Docs: https://problem_bank_scripts.readthedocs.io/en/latest/\n\n## Contributors\n\nWe welcome and recognize all contributions. You can see a list of current contributors in the [contributors tab](https://github.com/open-resources/problem_bank_scripts/graphs/contributors).\n\n### Credits\n\nThis package was created with Cookiecutter and the UBC-MDS/cookiecutter-ubc-mds project template, modified from the [pyOpenSci/cookiecutter-pyopensci](https://github.com/pyOpenSci/cookiecutter-pyopensci) project template and the [audreyr/cookiecutter-pypackage](https://github.com/audreyr/cookiecutter-pypackage).\n',
    'author': 'Open Problem Bank Team',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/open-resources/problem_bank_scripts',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# This file has been copied directly from the PL repo: https://github.com/PrairieLearn/PrairieLearn/blob/master/lib/python_helper_sympy.py

import sympy
# import ast
# import sys

# import lxml.html
# import html
# import to_precision
import numpy as np
# import uuid
# import sympy
import pandas
# from python_helper_sympy import convert_string_to_sympy
# from python_helper_sympy import sympy_to_json
# from python_helper_sympy import json_to_sympy
# import re
# import colors
# import unicodedata
# import importlib
# import importlib.util
# import os
import collections

# Create a new instance of this class to access the member dictionaries. This
# is to avoid accidentally modifying these dictionaries.
class _Constants:
    def __init__(self):
        self.helpers = {
            '_Integer': sympy.Integer,
        }
        self.variables = {
            'pi': sympy.pi,
            'e': sympy.E,
        }
        self.hidden_variables = {
            '_Exp1': sympy.E,
        }
        self.complex_variables = {
            'i': sympy.I,
            'j': sympy.I,
        }
        self.hidden_complex_variables = {
            '_ImaginaryUnit': sympy.I,
        }
        self.functions = {
            # These are shown to the student
            'cos': sympy.cos,
            'sin': sympy.sin,
            'tan': sympy.tan,
            'arccos': sympy.acos,
            'arcsin': sympy.asin,
            'arctan': sympy.atan,
            'acos': sympy.acos,
            'asin': sympy.asin,
            'atan': sympy.atan,
            'arctan2': sympy.atan2,
            'atan2': sympy.atan2,
            'exp': sympy.exp,
            'log': sympy.log,
            'sqrt': sympy.sqrt,
        }

# # Safe evaluation of user input to convert from string to sympy expression.
# #
# # Adapted from:
# # https://stackoverflow.com/a/30516254
# #
# # Documentation of ast:
# # https://docs.python.org/3/library/ast.html
# #
# # More documentation of ast:
# # https://greentreesnakes.readthedocs.io/
# #
# # FIXME: As of 2017-08-27 there is an open sympy issue discussing a
# # similar approach: https://github.com/sympy/sympy/issues/10805 and an
# # associated PR: https://github.com/sympy/sympy/pull/12524 but it is
# # unclear when/if they will be merged. We should check once sympy 1.2
# # is released and see whether we can switch to using
# # `sympify(..., safe=True)`.
# #
# # For examples of the type of attacks that we are avoiding:
# # https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html
# # http://blog.delroth.net/2013/03/escaping-a-python-sandbox-ndh-2013-quals-writeup/
# #
# # Another class of attacks is those that try and consume excessive
# # memory or CPU (e.g., `10**100**100`). We handle these attacks by the
# # fact that the entire element code runs in an isolated subprocess, so
# # this type of input will be caught and treated as a question code
# # failure.
# #
# # Other approaches for safe(r) eval in Python are:
# #
# # PyParsing:
# #     http://pyparsing.wikispaces.com
# #     http://pyparsing.wikispaces.com/file/view/fourFn.py
# #
# # RestrictedPython:
# #     https://pypi.python.org/pypi/RestrictedPython
# #     http://restrictedpython.readthedocs.io/
# #
# # Handling timeouts explicitly:
# #     http://code.activestate.com/recipes/496746-restricted-safe-eval/
# #
# # A similar (but more complex) approach to us:
# #     https://github.com/newville/asteval

# class Error(Exception):
#     def __init__(self, offset):
#         self.offset = offset

# class HasFloatError(Error):
#     def __init__(self, offset, n):
#         super(HasFloatError, self).__init__(offset)
#         self.n = n

# class HasComplexError(Error):
#     def __init__(self, offset, n):
#         super(HasComplexError, self).__init__(offset)
#         self.n = n

# class HasInvalidExpressionError(Error):
#     pass

# class HasInvalidFunctionError(Error):
#     def __init__(self, offset, text):
#         super(HasInvalidFunctionError, self).__init__(offset)
#         self.text = text

# class HasInvalidVariableError(Error):
#     def __init__(self, offset, text):
#         super(HasInvalidVariableError, self).__init__(offset)
#         self.text = text

# class HasParseError(Error):
#     pass

# class HasEscapeError(Error):
#     pass

# class HasCommentError(Error):
#     pass

# class CheckNumbers(ast.NodeTransformer):
#     def visit_Num(self, node):
#         if isinstance(node.n, int):
#             return ast.Call(func=ast.Name(id='_Integer', ctx=ast.Load()), args=[node], keywords=[])
#         elif isinstance(node.n, float):
#             raise HasFloatError(node.col_offset, node.n)
#         elif isinstance(node.n, complex):
#             raise HasComplexError(node.col_offset, node.n)
#         return node

# class CheckWhiteList(ast.NodeVisitor):
#     def __init__(self, whitelist):
#         self.whitelist = whitelist

#     def visit(self, node):
#         if not isinstance(node, self.whitelist):
#             node = get_parent_with_location(node)
#             raise HasInvalidExpressionError(node.col_offset)
#         return super().visit(node)

# class CheckFunctions(ast.NodeVisitor):
#     def __init__(self, functions):
#         self.functions = functions
#     def visit_Call(self, node):
#         if isinstance(node.func, ast.Name):
#             if node.func.id not in self.functions:
#                 node = get_parent_with_location(node)
#                 raise HasInvalidFunctionError(node.col_offset, node.func.id)
#         self.generic_visit(node)

# class CheckVariables(ast.NodeVisitor):
#     def __init__(self, variables):
#         self.variables = variables
#     def visit_Name(self, node):
#         if isinstance(node.ctx, ast.Load):
#             if not is_name_of_function(node):
#                 if node.id not in self.variables:
#                     node = get_parent_with_location(node)
#                     raise HasInvalidVariableError(node.col_offset, node.id)
#         self.generic_visit(node)

# def is_name_of_function(node):
#     # The node is the name of a function if all of the following are true:
#     # 1) it has type ast.Name
#     # 2) its parent has type ast.Call
#     # 3) it is not in the list of parent's args
#     return isinstance(node, ast.Name) and isinstance(node.parent, ast.Call) and (node not in node.parent.args)

# def get_parent_with_location(node):
#     if hasattr(node, 'col_offset'):
#         return node
#     else:
#         return get_parent_with_location(node.parent)

# def evaluate(expr, locals_for_eval={}):

#     # Disallow escape character
#     if '\\' in expr:
#         raise HasEscapeError(expr.find('\\'))

#     # Disallow comment character
#     if '#' in expr:
#         raise HasCommentError(expr.find('#'))

#     # Parse (convert string to AST)
#     try:
#         root = ast.parse(expr, mode='eval')
#     except Exception as err:
#         raise HasParseError(err.offset)

#     # Link each node to its parent
#     for node in ast.walk(root):
#         for child in ast.iter_child_nodes(node):
#             child.parent = node

#     # Disallow functions that are not in locals_for_eval
#     CheckFunctions(locals_for_eval['functions']).visit(root)

#     # Disallow variables that are not in locals_for_eval
#     CheckVariables(locals_for_eval['variables']).visit(root)

#     # Disallow AST nodes that are not in whitelist
#     #
#     # Be very careful about adding to the list below. In particular,
#     # do not add `ast.Attribute` without fully understanding the
#     # reflection-based attacks described by
#     # https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html
#     # http://blog.delroth.net/2013/03/escaping-a-python-sandbox-ndh-2013-quals-writeup/
#     #
#     whitelist = (ast.Module, ast.Expr, ast.Load, ast.Expression, ast.Call, ast.Name, ast.Num, ast.UnaryOp, ast.UAdd, ast.USub, ast.BinOp, ast.Add, ast.Sub, ast.Mult, ast.Div, ast.Mod, ast.Pow)
#     CheckWhiteList(whitelist).visit(root)

#     # Disallow float and complex, and replace int with sympy equivalent
#     root = CheckNumbers().visit(root)

#     # Clean up lineno and col_offset attributes
#     ast.fix_missing_locations(root)

#     # Convert AST to code and evaluate it with no global expressions and with
#     # a whitelist of local expressions
#     locals = {}
#     for key in locals_for_eval:
#         locals = {**locals, **locals_for_eval[key]}
#     return eval(compile(root, '<ast>', 'eval'), {'__builtins__': None}, locals)

# def convert_string_to_sympy(a, variables, allow_hidden=False, allow_complex=False):
#     const = _Constants()

#     # Create a whitelist of valid functions and variables (and a special flag
#     # for numbers that are converted to sympy integers).
#     locals_for_eval = {
#         'functions': const.functions,
#         'variables': const.variables,
#         'helpers': const.helpers,
#     }
#     if allow_hidden:
#         locals_for_eval['variables'] = {**locals_for_eval['variables'], **const.hidden_variables}
#     if allow_complex:
#         locals_for_eval['variables'] = {**locals_for_eval['variables'], **const.complex_variables}
#         if allow_hidden:
#             locals_for_eval['variables'] = {**locals_for_eval['variables'], **const.hidden_complex_variables}

#     # If there is a list of variables, add each one to the whitelist
#     if variables is not None:
#         for variable in variables:
#             locals_for_eval['variables'][variable] = sympy.Symbol(variable)

#     # Do the conversion
#     return evaluate(a, locals_for_eval)

# def point_to_error(s, ind, w = 5):
#     w_left = ind - max(0, ind-w)
#     w_right = min(ind+w, len(s)) - ind
#     return s[ind-w_left:ind+w_right] + '\n' + ' '*w_left + '^' + ' '*w_right

def sympy_to_json(a, allow_complex=True):
    const = _Constants()

    # Get list of variables in the sympy expression
    variables = [str(v) for v in a.free_symbols]

    # Check that variables do not conflict with reserved names
    reserved = {**const.helpers, **const.variables, **const.hidden_variables, **const.functions}
    if allow_complex:
        reserved = {**reserved, **const.complex_variables, **const.hidden_complex_variables}
    for k in reserved.keys():
        for v in variables:
            if k == v:
                raise ValueError('sympy expression has a variable with a reserved name: {:s}'.format(k))

    # Apply substitutions for hidden variables
    a = a.subs([(const.hidden_variables[key], key) for key in const.hidden_variables.keys()])
    if allow_complex:
        a = a.subs([(const.hidden_complex_variables[key], key) for key in const.hidden_complex_variables.keys()])

    return {'_type': 'sympy', '_value': str(a), '_variables': variables}

# def json_to_sympy(a, allow_complex=True):
#     if not '_type' in a:
#         raise ValueError('json must have key _type for conversion to sympy')
#     if a['_type'] != 'sympy':
#         raise ValueError('json must have _type == "sympy" for conversion to sympy')
#     if not '_value' in a:
#         raise ValueError('json must have key _value for conversion to sympy')
#     if not '_variables' in a:
#         a['_variables'] = None

#     return convert_string_to_sympy(a['_value'], a['_variables'], allow_hidden=True, allow_complex=allow_complex)

# Added to_json() from this file: https://github.com/PrairieLearn/PrairieLearn/blob/master/question-servers/freeformPythonLib/prairielearn.py
def to_json(v):
    """to_json(v)
    If v has a standard type that cannot be json serialized, it is replaced with
    a {'_type':..., '_value':...} pair that can be json serialized:
        complex -> '_type': 'complex'
        non-complex ndarray (assumes each element can be json serialized) -> '_type': 'ndarray'
        complex ndarray -> '_type': 'complex_ndarray'
        sympy.Expr (i.e., any scalar sympy expression) -> '_type': 'sympy'
        sympy.Matrix -> '_type': 'sympy_matrix'
    If v is an ndarray, this function preserves its dtype (by adding '_dtype' as
    a third field in the dictionary).
    This function does not try to preserve information like the assumptions on
    variables in a sympy expression.
    If v can be json serialized or does not have a standard type, then it is
    returned without change.
    """
    if np.isscalar(v) and np.iscomplexobj(v):
        return {'_type': 'complex', '_value': {'real': v.real, 'imag': v.imag}}
    elif isinstance(v, np.ndarray):
        if np.isrealobj(v):
            return {'_type': 'ndarray', '_value': v.tolist(), '_dtype': str(v.dtype)}
        elif np.iscomplexobj(v):
            return {'_type': 'complex_ndarray', '_value': {'real': v.real.tolist(), 'imag': v.imag.tolist()}, '_dtype': str(v.dtype)}
    elif isinstance(v, sympy.Expr):
        return sympy_to_json(v)
    elif isinstance(v, sympy.Matrix) or isinstance(v, sympy.ImmutableMatrix):
        s = [str(a) for a in v.free_symbols]
        num_rows, num_cols = v.shape
        M = []
        for i in range(0, num_rows):
            row = []
            for j in range(0, num_cols):
                row.append(str(v[i, j]))
            M.append(row)
        return {'_type': 'sympy_matrix', '_value': M, '_variables': s, '_shape': [num_rows, num_cols]}
    elif isinstance(v, pandas.DataFrame):
        return {'_type': 'dataframe', '_value': {'index': list(v.index), 'columns': list(v.columns), 'data': v.values.tolist()}}
    else:
        return v